# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `wehjit.application` module.
"""

from assetslib import Assets
from helpers import raises, assert_equal
from wehjit import application, errors, styles, base
from wehjit.collection import builtins, Collection


class test_Application(object):
    """
    Test the `application.Application` class.
    """
    klass = application.Application

    def test_init(self):
        """
        Test the `application.Application.__init__` method.
        """
        # Test with no args:
        inst = self.klass()
        assert inst.url == '/'
        assert isinstance(inst.assets, Assets)
        assert inst.assets.url == '/_/'
        assert inst.assets.prod is True
        assert inst.widgets is builtins
        assert inst.prod is True
        assert isinstance(inst.styles, styles.Styles)
        assert list(inst) == []

        # Test the plugins and pages methods:
        assert list(inst.plugins()) == []
        assert list(inst.pages()) == []

        # Test when prod=False
        inst = self.klass(prod=False)
        assert inst.prod is False
        assert inst.assets.prod is True
        assert inst.assets.url == '/_/'

        # Test with args:
        widgets = Collection()
        assets = Assets()
        inst = self.klass('/my-app', assets, widgets, False)
        assert inst.url == '/my-app/'
        assert inst.assets is assets
        assert inst.widgets is widgets
        assert inst.prod is False

        # Test with kwargs:
        inst = self.klass(url='/my-app///', assets=assets, widgets=widgets,
            prod=False)
        assert inst.url == '/my-app/'
        assert inst.assets is assets
        assert inst.widgets is widgets
        assert inst.prod is False

        # Test some more url variations:
        inst = self.klass(url='')
        assert inst.url == ''
        assert inst.assets.url == '_/'

        inst = self.klass(url=None)
        assert inst.url == ''
        assert inst.assets.url == '_/'

        inst = self.klass(url='some/relative/url')
        assert inst.url == 'some/relative/url/'
        assert inst.assets.url == '_/'

        inst = self.klass(url='another/relative////')
        assert inst.url == 'another/relative/'
        assert inst.assets.url == '_/'

    def test_new(self):
        """
        Test the `application.Application.new` method.
        """
        inst = self.klass(url='/my-app/')

        b = inst.new('Button', name='foo')
        assert isinstance(b, builtins.Button)
        assert b.app is inst
        assert b.name == 'foo'
        assert b.id is None
        assert list(inst.plugins()) == [builtins.Button]
        assert list(inst.pages()) == []

        p = inst.new('Page', id='foo:bar')
        assert isinstance(p, builtins.Page)
        assert p.app is inst
        assert p.id == 'foo:bar'
        assert p.baseurl == '/my-app/'
        assert p.url == '/my-app/foo/bar'
        assert p.static is False
        assert list(inst.plugins()) == [
            builtins[name] for name in ['Button', 'Page']
        ]
        assert list(inst.pages()) == [p]

        # id-map doesn't get built till Application.finalize()
        assert list(inst) == []

        # Test with static=True
        inst = self.klass(url='/foo/', static=True)
        p = inst.new('Page', id='bar:baz')
        assert p.app is inst
        assert p.id == 'bar:baz'
        assert p.baseurl == '/foo/'
        assert p.url == '/foo/bar/baz/index.xhtml'
        assert p.static is True

        p = inst.new('Page', id='')
        assert p.app is inst
        assert p.id == ''
        assert p.baseurl == '/foo/'
        assert p.url == '/foo/index.xhtml'
        assert p.static is True

    def test_finalize(self):
        """
        Test the `application.Application.finalize` method.
        """
        inst = self.klass(url='/my-app/')
        p1 = inst.new('Page', id='foo:bar')
        p2 = inst.new('Page', id='foo:bar')
        p2.id = 'bar:foo'
        assert list(inst) == []
        inst.finalize()
        assert list(inst) == ['bar:foo', 'foo:bar']
        assert inst['bar:foo'] is p2
        assert inst['foo:bar'] is p1

        # Test with conflicting page ids:
        inst = self.klass(url='/my-app/')
        p1 = inst.new('Page', id='foo:bar')
        p2 = inst.new('Page', id='bar:foo')
        p2.id = 'foo:bar'
        assert list(inst) == []
        e = raises(errors.PageConflict, inst.finalize)
        assert e.new is p2
        assert e.orig is p1
        assert e.app is inst
