# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `wehjit.base` module.
"""

from helpers import raises, assert_equal
from wehjit.constants import TYPE_ERROR, OVERRIDE_ERROR
from wehjit.freezable import freeze
from wehjit import base, errors


def test_relurl():
    f = base.relurl

    assert f('/_/site.css', '/reports/daily') == '/_/site.css'
    assert f('_/site.css', '/reports/daily') == '_/site.css'
    assert f('/_/site.css', 'reports/daily') == '/_/site.css'
    assert f('_/site.css', 'reports/daily') == '../../_/site.css'

    assert f('/_/site.css', '/reports/daily/') == '/_/site.css'
    assert f('_/site.css', '/reports/daily/') == '_/site.css'
    assert f('/_/site.css', 'reports/daily/') == '/_/site.css'
    assert f('_/site.css', 'reports/daily/') == '../../_/site.css'

    assert f('/_/site.css', '/') == '/_/site.css'
    assert f('_/site.css', '/') == '_/site.css'
    assert f('/_/site.css', '') == '/_/site.css'
    assert f('_/site.css', '') == '_/site.css'


class test_Widget(object):
    """
    Test the `base.Widget` class.
    """

    klass = base.Widget

    def test_init(self):
        """
        Test the `base.Widget.__init__` method.
        """
        inst = self.klass()
        assert inst.__request__ == dict()
        assert inst.__state__ is None

        # Test the descriptor keys on class and instance:
        keys = ('app', 'id', 'name')
        assert_equal(self.klass.Static, keys, check_type=True)
        assert_equal(inst.Static, keys, check_type=True)

        keys = tuple()
        assert_equal(self.klass.Dynamic, keys, check_type=True)
        assert_equal(inst.Dynamic, keys, check_type=True)

        keys = ('klass', 'relurl', 'tag')
        assert_equal(self.klass.StaticProp, keys, check_type=True)
        assert_equal(inst.StaticProp, keys, check_type=True)

        keys = ('css_classes',)
        assert_equal(self.klass.DynamicProp, keys, check_type=True)
        assert_equal(inst.DynamicProp, keys, check_type=True)


        # Static
        assert inst.name is None
        assert inst.id is None
        assert inst.app is None


        # StaticProp
        assert inst.klass == 'Widget'
        assert inst.tag is None

        # DynamicProp:
        assert inst.css_classes == 'Widget'
        inst.extra_css_classes = ('zero', 'Widget', 'one', 'Widget')
        assert inst.css_classes == 'Widget one zero'
        inst.extra_css_classes = tuple()
        assert inst.css_classes == 'Widget'
        return
        # Now lets set some stuff:
        items = dict(
            name='bar',
            id='foo.bar',
            baseurl='/example/',
            debug=True,
            title='Gone with the wind',
        )
        state = dict(items)
        state.update(
            dict(
                app=None,
                klass='Widget',
                tag=None,
                url=None,
            )
        )
        state = tuple(
            (key, state[key]) for key in sorted(state)
        )

        for (key, value) in items.iteritems():
            setattr(inst, key, value)
        for (key, value) in items.iteritems():
            assert getattr(inst, key) is value
            assert getattr(inst, '_' + key) is value
        assert inst.__state__ is None
        assert inst.__request__ == dict()

        # How about if we freeze it?
        freeze(inst)
        assert inst.__frozen__ is True
        inst.title = 'Dirty Harry'
        assert inst.__state__ == state
        assert inst.__request__ == dict(title='Dirty Harry')

        # Lets try again, but passing **kw to __init__():
        inst = self.klass(**items)
        for (key, value) in items.iteritems():
            assert getattr(inst, key) is value
            assert getattr(inst, '_' + key) is value
        assert inst.__state__ == None
        assert inst.__request__ == dict()

        # And we'll freeze it again:
        freeze(inst)
        assert inst.__frozen__ is True
        inst.title = 'Dirty Harry'
        assert inst.__state__ == state
        assert inst.__request__ == dict(title='Dirty Harry')

    def test_repr(self):
        """
        Test the `base.Widget.__repr__` method.
        """
        inst = self.klass(id='foo.bar', name='bar')
        assert repr(inst) == "Widget(id='foo.bar', name='bar')"

    def test_new(self):
        """
        Test the `base.Widget.new` method.
        """
        inst = self.klass()
        e = raises(errors.MissingApplication, inst.new, 'Field', id='foo', name='bar')
        assert e.klass == 'Field'
        assert e.widget is inst

        class App(object):
            @classmethod
            def new(cls, widget, **kw):
                return ('Yup', widget, kw)

        inst.app = App
        assert inst.new('Field', id='foo', name='bar') == \
            ('Yup', 'Field', dict(id='foo', name='bar'))

    def test_parent(self):
        """
        Test the `base.Widget.parent` property.
        """
        inst = self.klass()
        assert inst.parent is None
        p1 = base.Container(name='parent')
        inst.parent = p1
        assert inst.parent is p1
        p2 = base.Container(name='another_parent')
        e = raises(AttributeError, setattr, inst, 'parent', p2)
        assert str(e) == OVERRIDE_ERROR % ('Widget', 'parent', p1, p2)

        # Test with wrong type
        child = self.klass()
        parent = self.klass()
        e = raises(TypeError, setattr, child, 'parent', parent)
        assert str(e) == TYPE_ERROR % (
            'Widget.parent', base.Container, self.klass, parent
        )

    def test_generate(self):
        """
        Test the `base.Widget.generate` method.
        """
        # With no Widget.xml template, generate should return None:
        inst = self.klass()
        assert inst.generate() is None

    def test_render(self):
        """
        Test the `base.Widget.render` method.
        """
        # With no Widget.xml template, render should return '':
        inst = self.klass()
        assert inst.render() == ''


class test_Container(object):
    """
    Test the `base.Container` class.
    """
    klass = base.Container

    def test_add(self):
        """
        Test the `base.Container.add` method.
        """
        # Test with the wrong type:
        inst = self.klass()
        wrong = 'Wrong!'
        e = raises(TypeError, inst.add, wrong)
        assert str(e) == TYPE_ERROR % ('child', base.Widget, str, wrong)
        page = base.Page()
        e = raises(TypeError, inst.add, page)
        assert str(e) == 'child: a %r instance cannot be a child' % base.Page

        # Test with the correct type:
        okay = base.Widget()
        assert okay.parent is None
        inst.add(okay)
        assert okay.parent is inst
        assert okay.name is None
        assert okay.id is None
        assert list(inst.iter()) == [okay]

    def test_finalize(self):
        """
        Test the `base.Container.finalize` method.
        """
        # Test the stable anonymous name assignment:
        class Foo(base.Widget):
            pass
        class Bar(self.klass):
            pass
        inst = self.klass(id='baz')
        (foo1, foo2, foo3) = (Foo(name='hello'), Foo(), Foo())
        (bar1, bar2, bar3) = (Bar(), Bar(name='world'), Bar())
        form = base.Form()
        field = base.Field()
        children = (foo1, bar1, foo2, foo3, bar2, bar3, form, field)
        inst.add(*children)
        for c in children:
            assert c.name in (None, 'hello', 'world')
            assert c.id is None
        inst.finalize()

        assert foo1.name == 'hello'
        assert foo2.name == 'foo2'
        assert foo3.name == 'foo3'
        assert bar1.name == 'bar1'
        assert bar2.name == 'world'
        assert bar3.name == 'bar3'
        assert form.name is None
        assert field.name is None

        assert foo1.id == 'baz.hello'
        assert foo2.id == 'baz.foo2'
        assert foo3.id == 'baz.foo3'
        assert bar1.id == 'baz.bar1'
        assert bar2.id == 'baz.world'
        assert bar3.id == 'baz.bar3'
        assert form.id is None
        assert field.id is None

        # Test name conflicts:
        inst = self.klass()
        child1 = base.Widget(name='foo')
        child2 = base.Container(name='foo')
        inst.add(child1, child2)
        e = raises(errors.NameConflict, inst.finalize)
        assert e.new is child2
        assert e.orig is child1
        assert e.parent is inst

        inst = self.klass()
        child1 = base.Widget(name='foo')
        child2 = base.Container(name='bar')
        inst.add(child1, child2)
        child2.name = 'foo'
        e = raises(errors.NameConflict, inst.finalize)
        assert e.new is child2
        assert e.orig is child1
        assert e.parent is inst

    def test_iter(self):
        """
        Test the `base.Container.iter` method.
        """
        inst = self.klass()
        assert list(inst.iter()) == []
        children = [base.Widget() for i in xrange(10)]
        inst.add(*children)
        assert list(inst.iter()) == children

    def test_deepiter(self):
        """
        Test the `base.Container.deepiter` method.
        """
        c1 = self.klass()
        c2 = self.klass()
        c3 = self.klass()
        w1 = base.Widget()
        w2 = base.Widget()
        w3 = base.Widget()
        w4 = base.Widget()
        c1.add(w1, c2)
        c2.add(c3, w2, w3)
        c3.add(w4)
        assert list(c3.deepiter()) == [w4]
        assert list(c2.deepiter()) == [c3, w4, w2, w3]
        assert list(c1.deepiter()) == [w1, c2, c3, w4, w2, w3]
        c1.finalize()
        assert list(c3.deepiter()) == [w4]
        assert list(c2.deepiter()) == [c3, w4, w2, w3]
        assert list(c1.deepiter()) == [w1, c2, c3, w4, w2, w3]


class test_Form(object):
    """
    Test the `base.Form` class.
    """

    klass = base.Form

    def test_init(self):
        """
        Test the `base.Form.__init__` method.
        """
        inst = self.klass()
        assert inst.action is None
        assert inst.method == 'POST'
        inst = self.klass(action='https://whatever.com/stuff', method='GET')
        inst.action = 'https://whatever.com/stuff'
        inst.method = 'GET'

    def test_finalize(self):
        """
        Test the `base.Form.finalize` method.
        """
        inst = self.klass()
        w = base.Widget()
        c = base.Container()
        f1 = base.Field()
        f2 = base.Field()
        f3 = base.Field()
        c.add(f1, f2)
        inst.add(w, c, f3)
        for x in (w, c, f1, f2, f3):
            assert x.name is None
            assert x.id is None
        inst.finalize()
        assert f1.name == 'field1'
        f2.name == 'field2'
        f3.name == 'field3'

        # Test that FieldConflict gets raised:
        inst = self.klass()
        w = base.Widget()
        c = base.Container()
        f1 = base.Field(name='first')
        f2 = base.Field(name='second')
        f3 = base.Field(name='first')
        c.add(f1, f2)
        inst.add(w, c, f3)
        e = raises(errors.FieldConflict, inst.finalize)
        assert e.new is f3
        assert e.orig is f1
        assert e.form is inst


class test_Field(object):
    """
    Test the `base.Field` class.
    """

    klass = base.Field

    def test_form(self):
        """
        Test the `base.Field.form` property.
        """
        inst = self.klass()
        assert inst.form is None
        f1 = base.Form(id='foo')
        inst.form = f1
        assert inst.form is f1
        f2 = base.Form(id='bar')
        e = raises(AttributeError, setattr, inst, 'form', f2)
        assert str(e) == OVERRIDE_ERROR % ('Field', 'form', f1, f2)

        # Test with wrong type
        field = self.klass()
        form = base.Container()
        e = raises(TypeError, setattr, field, 'form', form)
        assert str(e) == TYPE_ERROR % (
            'Field.form', base.Form, base.Container, form
        )

    def test_finalize(self):
        """
        Test the `base.Field.finalize` method.

        There isn't a specific `base.Field.finalize()` method, so what this
        really tests is that `base.Widget.finalize()` behaves appropriately
        with regard to the `base.Field.base_id_from` class attribute (whose
        value is ``'form'``).
        """
        # Test when form is None:
        inst = self.klass(name='a_field')
        assert inst.id is None
        inst.finalize()
        assert inst.id == 'a_field'

        # Test when form.id is None:
        inst = self.klass(name='a_field')
        form = base.Form()
        inst.form = form
        assert inst.id is None
        inst.finalize()
        assert inst.id == 'a_field'

        # Test when form.id is not None:
        inst = self.klass(name='a_field')
        form = base.Form(id='the_form')
        inst.form = form
        assert inst.id is None
        inst.finalize()
        assert inst.id == 'the_form.a_field'


class test_Page(object):
    """
    Test the `base.Page` class.
    """
    klass = base.Page

    def test_init(self):
        """
        Test the `base.Page.__init__` method.
        """
        inst = self.klass()
        assert inst.__request__ == dict()
        assert inst.__state__ is None

        # Test the descriptor keys on class and instance:
        keys = (
            'app', 'baseurl', 'id', 'inline_css', 'inline_js', 'serializer',
            'static',
        )
        assert_equal(self.klass.Static, keys, check_type=True)
        assert_equal(inst.Static, keys, check_type=True)

        keys = ('lang', 'title')
        assert_equal(self.klass.Dynamic, keys, check_type=True)
        assert_equal(inst.Dynamic, keys, check_type=True)

        keys = (
            'children', 'content_type', 'doctype', 'klass', 'relurl', 'tag',
            'url', 'urldir',
        )
        assert_equal(self.klass.StaticProp, keys, check_type=True)
        assert_equal(inst.StaticProp, keys, check_type=True)

        keys = ('css_classes', 'links_css', 'links_js')
        assert_equal(self.klass.DynamicProp, keys, check_type=True)
        assert_equal(inst.DynamicProp, keys, check_type=True)

        # Test default values of state properties:
        assert inst.app is None
        assert inst.baseurl == ''
        assert inst.id == ''
        assert inst.serializer == 'xhtml'
        assert inst.static is False

        assert inst.lang == 'en'
        assert inst.title == 'Page Title'

        assert inst.children == tuple()
        assert inst.content_type == 'application/xhtml+xml; charset=utf-8'
        assert inst.doctype == 'xhtml11'
        assert inst.inline_css is None
        assert inst.inline_js is None
        assert inst.klass == 'Page'
        assert inst.tag == 'html'
        assert inst.url == ''

        assert inst.css_classes == 'Page'
        assert inst.links_css == tuple()
        assert inst.links_js == tuple()

    def test_filename(self):
        """
        Test the `base.Page.filename` method.
        """

        # Test with serializer='xhtml':
        inst = self.klass(baseurl='woo/', id='foo:bar')
        assert inst.filename('/out/') == '/out/rel/woo/foo/bar/index.xhtml'
        assert inst.filename('out') == 'out/rel/woo/foo/bar/index.xhtml'

        inst = self.klass(baseurl='/woo/', id='foo:bar')
        assert inst.filename('/out/') == '/out/abs/woo/foo/bar/index.xhtml'
        assert inst.filename('out') == 'out/abs/woo/foo/bar/index.xhtml'

        inst = self.klass(baseurl='', id='foo:bar')
        assert inst.filename('/out/') == '/out/rel/foo/bar/index.xhtml'
        assert inst.filename('out') == 'out/rel/foo/bar/index.xhtml'

        inst = self.klass(baseurl='/', id='foo:bar')
        assert inst.filename('/out/') == '/out/abs/foo/bar/index.xhtml'
        assert inst.filename('out') == 'out/abs/foo/bar/index.xhtml'

        inst = self.klass(baseurl='woo/moo/', id='')
        assert inst.filename('/out/') == '/out/rel/woo/moo/index.xhtml'
        assert inst.filename('out') == 'out/rel/woo/moo/index.xhtml'

        inst = self.klass(baseurl='/woo/moo/', id='')
        assert inst.filename('/out/') == '/out/abs/woo/moo/index.xhtml'
        assert inst.filename('out') == 'out/abs/woo/moo/index.xhtml'

        inst = self.klass(baseurl='', id='')
        assert inst.filename('/out/') == '/out/rel/index.xhtml'
        assert inst.filename('out') == 'out/rel/index.xhtml'

        inst = self.klass(baseurl='/', id='')
        assert inst.filename('/out/') == '/out/abs/index.xhtml'
        assert inst.filename('out') == 'out/abs/index.xhtml'

        # Test with serializer='html':
        inst = self.klass(serializer='html', baseurl='woo/', id='foo:bar')
        assert inst.filename('/out/') == '/out/rel/woo/foo/bar/index.html'
        assert inst.filename('out') == 'out/rel/woo/foo/bar/index.html'

        inst = self.klass(serializer='html', baseurl='/woo/', id='foo:bar')
        assert inst.filename('/out/') == '/out/abs/woo/foo/bar/index.html'
        assert inst.filename('out') == 'out/abs/woo/foo/bar/index.html'

        inst = self.klass(serializer='html', baseurl='', id='foo:bar')
        assert inst.filename('/out/') == '/out/rel/foo/bar/index.html'
        assert inst.filename('out') == 'out/rel/foo/bar/index.html'

        inst = self.klass(serializer='html', baseurl='/', id='foo:bar')
        assert inst.filename('/out/') == '/out/abs/foo/bar/index.html'
        assert inst.filename('out') == 'out/abs/foo/bar/index.html'

        inst = self.klass(serializer='html', baseurl='woo/moo/', id='')
        assert inst.filename('/out/') == '/out/rel/woo/moo/index.html'
        assert inst.filename('out') == 'out/rel/woo/moo/index.html'

        inst = self.klass(serializer='html', baseurl='/woo/moo/', id='')
        assert inst.filename('/out/') == '/out/abs/woo/moo/index.html'
        assert inst.filename('out') == 'out/abs/woo/moo/index.html'

        inst = self.klass(serializer='html', baseurl='', id='')
        assert inst.filename('/out/') == '/out/rel/index.html'
        assert inst.filename('out') == 'out/rel/index.html'

        inst = self.klass(serializer='html', baseurl='/', id='')
        assert inst.filename('/out/') == '/out/abs/index.html'
        assert inst.filename('out') == 'out/abs/index.html'

    def test_urldir(self):
        """
        Test the `base.Page.urldir` StaticProp.
        """
        inst = self.klass(baseurl='', id='')
        assert inst.urldir == ''
        inst = self.klass(baseurl='/', id='')
        assert inst.urldir == '/'

        inst = self.klass(baseurl='', id='foo:bar')
        assert inst.urldir == 'foo/bar'
        inst = self.klass(baseurl='/', id='foo:bar')
        assert inst.urldir == '/foo/bar'

        inst = self.klass(baseurl='my-app/', id='')
        assert inst.urldir == 'my-app'
        inst = self.klass(baseurl='/my-app/', id='')
        assert inst.urldir == '/my-app'

        inst = self.klass(baseurl='my-app/', id='foo:bar')
        assert inst.urldir == 'my-app/foo/bar'
        inst = self.klass(baseurl='/my-app/', id='foo:bar')
        assert inst.urldir == '/my-app/foo/bar'

    def test_url(self):
        """
        Test the `base.Page.url` StaticProp.
        """
        # Test with id=''
        inst = self.klass(baseurl='/my-app/')
        assert inst.url == '/my-app/'
        inst = self.klass(baseurl='/my-app/', static=True)
        assert inst.url == '/my-app/index.xhtml'

        # Test with a multi-part id:
        inst = self.klass(baseurl='/my-app/', id='reports:daily:today')
        assert inst.url == '/my-app/reports/daily/today'
        inst = self.klass(baseurl='/my-app/', id='reports:daily:today',
                static=True)
        assert inst.url == '/my-app/reports/daily/today/index.xhtml'

        # Test all the above again but with serializer='html'
        inst = self.klass(baseurl='/my-app/', serializer='html')
        assert inst.url == '/my-app/'
        inst = self.klass(baseurl='/my-app/', static=True, serializer='html')
        assert inst.url == '/my-app/index.html'

        inst = self.klass(baseurl='/my-app/', id='reports:daily:today',
                serializer='html')
        assert inst.url == '/my-app/reports/daily/today'
        inst = self.klass(baseurl='/my-app/', id='reports:daily:today',
                static=True, serializer='html')
        assert inst.url == '/my-app/reports/daily/today/index.html'
