# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `wehjit.collection` module.
"""

from helpers import raises
from wehjit.constants import TYPE_ERROR
from wehjit import collection, base, errors


def get_plugins():
    """
    Return a tuple of Widget subclasses.
    """
    class Text(base.Widget):
        pass
    class Checkbox(base.Widget):
        pass
    class Radio(base.Widget):
        pass
    class SingleSelect(base.Widget):
        pass
    class MultiSelect(base.Widget):
        pass
    return (
        Text,
        Checkbox,
        Radio,
        SingleSelect,
        MultiSelect,
    )


class test_Collection(object):
    klass = collection.Collection

    def test_register(self):
        """
        Test the `collection.Collection.register` method.
        """
        # Test that TypeError is raised when not a class:
        inst = self.klass()
        Example = base.Widget()
        e = raises(TypeError, inst.register, Example)
        assert str(e) == 'plugin must be a class; got %r' % Example

        # Test that ValueError is raised when not a Widget subclass:
        class Example(object):
            pass
        e = raises(ValueError, inst.register, Example)
        assert str(e) ==  'plugin must be a Widget subclass; got %r' % Example

        # Test that OverrideError is raised when trying to implicitly override:
        class Example(base.Widget):
            pass
        inst.register(Example)
        eg1 = Example
        assert inst.Example is Example
        class Example(base.Widget):
            """
            This is a new class called ``Example``.
            """
        e = raises(errors.OverrideError, inst.register, Example)
        assert e.new is Example
        assert e.orig is eg1

        # but that explicit overriding works:
        inst.register(Example, override=True)
        assert inst.Example is Example

        # Test that MissingOverrideError is raised when trying to override a
        # plugin that doesn't exist:
        class Another(base.Widget):
            pass
        e = raises(errors.MissingOverrideError, inst.register, Another, override=True)
        assert e.name == 'Another'
        assert e.plugin is Another

        # but that `Another` can be registered:
        inst.register(Another)
        assert inst.Another is Another

        # Test that AttributeConflict is raised if plugin name conflicts with an
        # attribute on the registrar:
        class register(base.Widget):
            pass
        e = raises(errors.AttributeConflict, inst.register, register)
        assert e.plugin is register
        assert e.registrar == 'Collection'
        assert e.attribute == 'register'

        # but that a subclass of `register` can be registered:
        class Lastly(register):
            pass
        inst.register(Lastly)
        assert inst.Lastly is Lastly

    def test_len(self):
        """
        Test the `collection.Collection.__len__` method.
        """
        inst = self.klass()
        assert len(inst) == 0
        for (i, plugin) in enumerate(get_plugins()):
            inst.register(plugin)
            assert len(inst) == i + 1

    def test_iter(self):
        """
        Test the `collection.Collection.__iter__` method.
        """
        inst = self.klass()
        assert list(inst) == []
        registered = []
        for plugin in get_plugins():
            inst.register(plugin)
            registered.append(plugin)
            assert list(inst) == sorted(p.__name__ for p in registered)

    def test_contains(self):
        """
        Test the `collection.Collection.__contains__` method.
        """
        inst = self.klass()
        plugins = get_plugins()
        for plugin in plugins:
            assert plugin.__name__ not in inst
        for plugin in plugins:
            assert plugin.__name__ not in inst
            inst.register(plugin)
            assert plugin.__name__ in inst
        for plugin in plugins:
            assert plugin.__name__ in inst

    def test_getitem(self):
        """
        Test the `collection.Collection.__getitem__` method.
        """
        inst = self.klass()
        plugins = get_plugins()
        for plugin in plugins:
            e = raises(KeyError, inst.__getitem__, plugin.__name__)
            assert str(e) == repr(plugin.__name__)
        for plugin in plugins:
            e = raises(KeyError, inst.__getitem__, plugin.__name__)
            assert str(e) == repr(plugin.__name__)
            inst.register(plugin)
            assert inst[plugin.__name__] is plugin
        for plugin in plugins:
            assert inst[plugin.__name__] is plugin

    def test_todict(self):
        """
        Test the `collection.Collection.__todict__` method.
        """
        inst = self.klass()
        registered = dict()
        prev = None
        assert inst.__todict__() == dict()
        for plugin in get_plugins():
            inst.register(plugin)
            registered[plugin.__name__] = plugin
            d = inst.__todict__()
            assert d == registered
            assert d is not prev
            prev = d


def test_builtins():
    """
    Test the `collection.builtins` attribute.
    """
    assert type(collection.builtins) is collection.Collection
