# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `wehjit.freezable` module.
"""

from helpers import raises
from wehjit.constants import SET_ERROR, DEL_ERROR
from wehjit import freezable


class test_Freezable(object):
    """
    Test the `wehjit.freezable.Freezable` class
    """
    klass = freezable.Freezable

    def test_freeze(self):
        """
        Test the `wehjit.freezable.Freezable.__freeze__` method.
        """
        inst = self.klass()
        assert inst.__frozen__ is False
        inst.__freeze__()
        assert inst.__frozen__ is True

        # __freeze__() can only be called once:
        e = raises(StandardError, inst.__freeze__)
        assert str(e) == 'already frozen: %r' % inst
        assert inst.__frozen__ is True

    def test_setattr(self):
        """
        Test the `wehjit.freezable.Freezable.__setattr__` method.
        """
        inst = self.klass()
        inst.attr1 = 'Hello, world!'
        assert inst.attr1 == 'Hello, world!'
        inst.__frozen__ = True
        for name in ('attr1', 'attr2'):
            e = raises(AttributeError, setattr, inst, name, 'whatever')
            assert str(e) == SET_ERROR % ('Freezable', name, 'whatever')
        assert inst.attr1 == 'Hello, world!'

    def test_delattr(self):
        """
        Test the `wehjit.freezable.Freezable.__delattr__` method.
        """
        inst = self.klass()
        inst.attr1 = 'Hello, world!'
        inst.attr2 = 'How are you?'
        assert inst.attr1 == 'Hello, world!'
        assert inst.attr2 == 'How are you?'
        del inst.attr1
        assert not hasattr(inst, 'attr1')
        inst.__frozen__ = True
        e = raises(AttributeError, delattr, inst, 'attr2')
        assert str(e) == DEL_ERROR % ('Freezable', 'attr2')
        assert inst.attr2 == 'How are you?'


def test_freeze():
    """
    Test the `wehjit.freezable.freeze` function
    """
    f = freezable.freeze

    # Test with Freezable instance:
    inst = freezable.Freezable()
    assert inst.__frozen__ is False
    assert f(inst) is inst
    assert inst.__frozen__ is True
    e = raises(StandardError, f, inst)
    assert str(e) == 'already frozen: %r' % inst

    # Test with a class *without* __freeze__() method
    class Other(object):
        pass

    inst = Other()
    assert f(inst) is inst
    assert inst.__frozen__ is True
    e = raises(StandardError, f, inst)
    assert str(e) == 'already frozen: %r' % inst
    assert f(inst, check=False) is inst

    # Test with a class with a __freeze__() method:
    class Another(object):
        __called__ = False
        def __freeze__(self):
            self.__called__ = True
            self.__frozen__ = True

    inst = Another()
    assert f(inst) is inst
    assert inst.__called__ is True
    assert inst.__frozen__ is True
    e = raises(StandardError, f, inst)
    assert str(e) == 'already frozen: %r' % inst
    assert f(inst, check=False) is inst


def test_isfrozen():
    """
    Test the `wehjit.freezable.isfrozen` function.
    """
    f = freezable.isfrozen

    # Test with Freezable instance:
    inst = freezable.Freezable()
    assert f(inst) is False
    inst.__frozen__ = True
    assert f(inst) is True

    # Test with another class implemented freeze-protocol:
    class Another(object):
        pass

    inst = Another()
    assert f(inst) is False
    inst.__frozen__ = True
    assert f(inst) is True
