# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `wehjit.state` module.
"""

import threading
from helpers import raises, assert_equal
from wehjit.constants import TYPE_ERROR, CALLABLE_ERROR, SET_ERROR
from wehjit.freezable import freeze
from wehjit import state


class Dummy(object):
    def __init__(self):
        self.__frozen__ = False
        self.__threadlocal = threading.local()
        self.__request__ = self.__threadlocal.__dict__
        self.i = 0


class test_Descriptor(object):
    """
    Test the `state.Descriptor` class.
    """

    klass = state.Descriptor

    def test_init(self):
        """
        Test the `state.Descriptor.__init__` method.
        """
        # Test with wrong name type:
        e = raises(TypeError, self.klass, 18)
        assert str(e) == TYPE_ERROR % ('Descriptor.name', str, int, 18)

        # Test with wrong doc type:
        e = raises(TypeError, self.klass, 'foo', doc=18)
        assert str(e) == TYPE_ERROR % ('Descriptor.doc', str, int, 18)

        # Test with wrong json type:
        e = raises(TypeError, self.klass, 'foo', json=18)
        assert str(e) == TYPE_ERROR % ('Descriptor.json', bool, int, 18)

        # Test supplying only name:
        inst = self.klass('foo')
        assert inst.name == 'foo'
        assert inst.pname == '_foo'
        assert inst.doc is None
        assert inst.json is False

        # Test supplying all:
        inst = self.klass('foo', doc='the foo stops here', json=True)
        assert inst.name == 'foo'
        assert inst.pname == '_foo'
        assert inst.doc == 'the foo stops here'
        assert inst.json is True

    def test_repr(self):
        """
        Test the `state.Descriptor.__repr__` method.
        """
        inst = self.klass('foo')
        assert repr(inst) == "Descriptor('foo')"
        inst = self.klass('foo', json=False, doc=None)
        assert repr(inst) == "Descriptor('foo')"
        inst = self.klass('bar', json=True, doc='Hello naughty nurse!')
        assert repr(inst) == \
            "Descriptor('bar', doc='Hello naughty nurse!', json=True)"


class test_StateBase(object):
    """
    Test the `state.StateBase` class.
    """

    klass = state.StateBase

    def test_init(self):
        """
        Test the `state.StateBase.__init__` method.
        """
        e = raises(TypeError, self.klass, 18)
        assert str(e) == TYPE_ERROR % ('StateBase.name', str, int, 18)

        e = raises(TypeError, self.klass, 'hello', validator='world')
        assert str(e) == CALLABLE_ERROR % ('StateBase.validator', str, 'world')

        inst = self.klass('kind')
        assert inst.name == 'kind'
        assert inst.pname == '_kind'
        assert inst.validator is None
        assert inst.default is None
        assert inst.doc is None

        def valfoo():
            pass
        inst = self.klass('foo', validator=valfoo, default='bar', doc='Foo this')
        assert inst.name == 'foo'
        assert inst.pname == '_foo'
        assert inst.validator is valfoo
        assert inst.default == 'bar'
        assert inst.doc == 'Foo this'

    def test_repr(self):
        """
        Test the `state.StateBase.__repr__` method.
        """
        def valbar():
            pass
        inst = self.klass('bar', validator=valbar, default='BAR', doc='Doc this')
        assert repr(inst) == \
            "StateBase('bar', default='BAR', doc='Doc this', validator=valbar)"


class test_Static(object):
    """
    Test the `state.Static` class.
    """

    klass = state.Static

    def test_init(self):
        """
        Test the `state.Static.__init__` method.
        """
        e = raises(TypeError, self.klass, 18)
        assert str(e) == TYPE_ERROR % ('Static.name', str, int, 18)

        e = raises(TypeError, self.klass, 'hello', validator='world')
        assert str(e) == CALLABLE_ERROR % ('Static.validator', str, 'world')

        inst = self.klass('kind')
        assert inst.name == 'kind'
        assert inst.pname == '_kind'
        assert inst.validator is None
        assert inst.default is None
        assert inst.doc is None

        def valfoo():
            pass
        inst = self.klass('foo', validator=valfoo, default='bar', doc='Foo this')
        assert inst.name == 'foo'
        assert inst.pname == '_foo'
        assert inst.validator is valfoo
        assert inst.default == 'bar'
        assert inst.doc == 'Foo this'

    def test_repr(self):
        """
        Test the `state.Static.__repr__` method.
        """
        def valbar():
            pass
        inst = self.klass('bar',
            validator=valbar, default='BAR', doc='Doc this'
        )
        assert repr(inst) == \
            "Static('bar', default='BAR', doc='Doc this', validator=valbar)"

    def test_get(self):
        """
        Test the `state.Static.__get__` method.
        """
        class Example(Dummy):
            a = self.klass('a')

            def __validate_b(self, value):
                return value.upper()
            b = self.klass('b', validator=__validate_b, default='Be')

            def __validate_c(self, value):
                raise ValueError(
                    'No way, %s!' % (value,)
                )
            c = self.klass('c', validator=__validate_c, default='Sea')

        eg = Example()

        assert eg.a is None
        eg._a = 'aye'
        assert eg.a == 'aye'
        del eg._a
        assert eg.a is None

        assert eg.b == 'Be'
        eg._b = 'bee'
        assert eg.b == 'bee'
        del eg._b
        assert eg.b == 'Be'

        assert eg.c == 'Sea'
        eg._c = 'see'
        assert eg.c == 'see'
        del eg._c
        assert eg.c == 'Sea'

        # Test when frozen:
        eg._a = 'aye'
        eg._b = 'bee'
        eg._c = 'see'
        eg.__frozen__ = True
        assert eg.a == 'aye'
        assert eg.b == 'bee'
        assert eg.c == 'see'

        eg = Example()
        eg.__frozen__ = True
        assert eg.a is None
        assert eg.b == 'Be'
        assert eg.c == 'Sea'

    def test_set(self):
        """
        Test the `state.Static.__set__` method.
        """
        class Example(Dummy):
            a = self.klass('a')

            def __validate_b(self, value):
                return value.upper()
            b = self.klass('b', validator=__validate_b, default='Be')

            def __validate_c(self, value):
                raise ValueError(
                    'No way, %s!' % (value,)
                )
            c = self.klass('c', validator=__validate_c, default='Sea')

            d = self.klass('d', kind=int)

        eg = Example()

        assert eg.a is None
        assert not hasattr(eg, '_a')
        eg.a = 'aye'
        assert eg.a == 'aye'
        assert eg._a == 'aye'
        del eg._a
        assert eg.a is None

        assert eg.b == 'Be'
        assert not hasattr(eg, '_b')
        eg.b = 'bee'
        assert eg.b == 'BEE'
        assert eg._b == 'BEE'
        del eg._b
        assert eg.b == 'Be'

        assert eg.c == 'Sea'
        assert not hasattr(eg, '_c')
        e = raises(ValueError, setattr, eg, 'c', 'Jose')
        assert str(e) == 'No way, Jose!'
        assert not hasattr(eg, '_c')
        eg._c = 'see'
        assert eg.c == 'see'
        del eg._c
        assert eg.c == 'Sea'

        assert eg.d is None
        assert not hasattr(eg, '_d')
        e = raises(TypeError, setattr, eg, 'd', '18')
        assert str(e) == TYPE_ERROR % ('Example.d', int, str, '18')
        assert not hasattr(eg, '_d')
        eg.d = 18
        assert eg.d == 18
        assert eg._d == 18
        del eg._d
        assert eg.d is None

        # Test when frozen
        eg.a = 'aye'
        assert eg.a == 'aye'
        assert eg._a == 'aye'
        eg.b = 'bee'
        assert eg.b == 'BEE'
        assert eg._b == 'BEE'
        eg._c = 'see'
        eg.__frozen__ = True
        e = raises(AttributeError, setattr, eg, 'a', 'AA')
        assert str(e) == "frozen: cannot set Example.a to 'AA'"
        e = raises(AttributeError, setattr, eg, 'b', 'BB')
        assert str(e) == "frozen: cannot set Example.b to 'BB'"
        e = raises(ValueError, setattr, eg, 'c', 'CC')
        assert str(e) == 'No way, CC!'
        e = raises(TypeError, setattr, eg, 'd', '18')
        assert str(e) == TYPE_ERROR % ('Example.d', int, str, '18')

        eg = Example()
        eg.__frozen__ = True
        e = raises(AttributeError, setattr, eg, 'a', 'AA')
        assert str(e) == "frozen: cannot set Example.a to 'AA'"
        e = raises(AttributeError, setattr, eg, 'b', 'BB')
        assert str(e) == "frozen: cannot set Example.b to 'BB'"
        e = raises(ValueError, setattr, eg, 'c', 'CC')
        assert str(e) == 'No way, CC!'
        e = raises(TypeError, setattr, eg, 'd', '18')
        assert str(e) == TYPE_ERROR % ('Example.d', int, str, '18')


class test_Dynamic(object):
    """
    Test the `state.Dynamic` class.
    """

    klass = state.Dynamic

    def test_init(self):
        """
        Test the `state.Dynamic.__init__` method.
        """
        e = raises(TypeError, self.klass, 18)
        assert str(e) == TYPE_ERROR % ('Dynamic.name', str, int, 18)

        e = raises(TypeError, self.klass, 'hello', validator='world')
        assert str(e) == CALLABLE_ERROR % ('Dynamic.validator', str, 'world')

        inst = self.klass('kind')
        assert inst.name == 'kind'
        assert inst.pname == '_kind'
        assert inst.validator is None
        assert inst.default is None
        assert inst.doc is None

        def valfoo():
            pass
        inst = self.klass('foo', validator=valfoo, default='bar', doc='Foo this')
        assert inst.name == 'foo'
        assert inst.pname == '_foo'
        assert inst.validator is valfoo
        assert inst.default == 'bar'
        assert inst.doc == 'Foo this'

    def test_repr(self):
        """
        Test the `state.Dynamic.__repr__` method.
        """
        def valbar():
            pass
        inst = self.klass('bar',
            validator=valbar, default='BAR', doc='Doc this'
        )
        assert repr(inst) == \
            "Dynamic('bar', default='BAR', doc='Doc this', validator=valbar)"

    def test_get(self):
        """
        Test the `state.Dynamic.__get__` method.
        """
        class Example(Dummy):
            a = self.klass('a', validator=lambda: None)
            b = self.klass('b', default='Bee')
            c = self.klass('c', default='See')

        eg = Example()
        eg.__request__['a'] = 'A'
        eg.__request__['b'] = 'B'

        assert eg.a is None
        eg._a = 'aye'
        assert eg.a == 'aye'

        assert eg.b == 'Bee'
        eg._b = 'be'
        assert eg.b == 'be'

        assert eg.c == 'See'

        eg.__frozen__ = True
        assert eg.a == 'A'
        assert eg.b == 'B'
        assert eg.c == 'See'

        del eg.__request__['a']
        del eg.__request__['b']
        eg.__request__['c'] = 'sea'

        assert eg.a == 'aye'
        assert eg.b == 'be'
        assert eg.c == 'sea'

    def test_set(self):
        """
        Test the `state.Dynamic.__set__` method.
        """
        class Example(Dummy):
            a = self.klass('a', default='Aye')

            def _validate_b(self, value):
                return value.lower() * 2
            b = self.klass('b', validator=_validate_b)
            _b = 'Bee'

            def _validate_c(self, value):
                raise ValueError('no way, %s' % (value,))
            c = self.klass('c', validator=_validate_c, default='See')

        eg = Example()
        eg.__request__['a'] = 'yello'

        eg.a = 'hey'
        assert eg._a == 'hey'
        assert eg.__request__ == dict(a='yello')

        eg.b = 'MOIN'
        assert eg._b == 'moinmoin'
        assert eg.__request__ == dict(a='yello')

        e = raises(ValueError, setattr, eg, 'c', 'jose')
        assert str(e) == 'no way, jose'
        assert not hasattr(eg, '_c')
        assert eg.__request__ == dict(a='yello')

        # Now freeze():
        eg.__frozen__ = True

        eg.a = 'A'
        assert eg._a == 'hey'
        assert eg.__request__ == dict(a='A')

        eg.b = 'B'
        assert eg._b == 'moinmoin'
        assert eg.__request__ == dict(a='A', b='bb')

        e = raises(ValueError, setattr, eg, 'c', 'cuervo')
        assert str(e) == 'no way, cuervo'
        assert not hasattr(eg, '_c')
        assert eg.__request__ == dict(a='A', b='bb')


class test_PropBase(object):
    """
    Test the `state.PropBase` class.
    """

    klass = state.PropBase

    def test_init(self):
        """
        Test the `state.PropBase.__init__` method.
        """
        def getfoo():
            """
            My getfoo doc.
            """

        # Test with non-callable get:
        e = raises(TypeError, self.klass, 'hello', name='world')
        assert str(e) == CALLABLE_ERROR % ('PropBase.get', str, 'hello')

        # Test with non-str name:
        e = raises(TypeError, self.klass, name=False)
        assert str(e) == TYPE_ERROR % ('PropBase.name', str, bool, False)

        # Test providing neither get nor name:
        e = raises(TypeError, self.klass)
        assert str(e) == 'PropBase: must provide at least `get` or `name`'

        # Test providing get and default:
        e = raises(TypeError, self.klass, getfoo, default='Yo')
        assert str(e) == \
            "PropBase: cannot provide both get=%r and default='Yo'" % getfoo, str(e)

        # get
        inst = self.klass(getfoo)
        assert inst.get is getfoo
        assert inst.name == 'getfoo'
        assert inst.pname == '_getfoo'
        assert inst.default is None
        assert inst.doc == 'My getfoo doc.'

        # get + name
        inst = self.klass(getfoo, name='bar')
        assert inst.get is getfoo
        assert inst.name == 'bar'
        assert inst.pname == '_bar'
        assert inst.default is None
        assert inst.doc == 'My getfoo doc.'

        # get + name + doc
        inst = self.klass(getfoo, name='bar', doc='My bar doc.')
        assert inst.get is getfoo
        assert inst.name == 'bar'
        assert inst.pname == '_bar'
        assert inst.default is None
        assert inst.doc == 'My bar doc.'

        # name
        inst = self.klass(name='bar')
        assert inst.get is None
        assert inst.name == 'bar'
        assert inst.pname == '_bar'
        assert inst.default is None
        assert inst.doc is None

        # name + default
        inst = self.klass(name='bar', default='Baz')
        assert inst.get is None
        assert inst.name == 'bar'
        assert inst.pname == '_bar'
        assert inst.default == 'Baz'
        assert inst.doc is None

        # name + default + doc
        inst = self.klass(name='bar', default='Baz', doc='My bar doc.')
        assert inst.get is None
        assert inst.name == 'bar'
        assert inst.pname == '_bar'
        assert inst.default == 'Baz'
        assert inst.doc == 'My bar doc.'

    def test_repr(self):
        """
        Test the `state.PropBase.__repr__` method.
        """
        def getfoo():
            """
            My foodoc.
            """
            pass

        # get
        inst = self.klass(getfoo)
        assert repr(inst) == "PropBase(getfoo, 'getfoo', doc='My foodoc.')"

        # get + name + doc
        inst = self.klass(getfoo, 'bar', doc='My bardoc.')
        assert repr(inst) == "PropBase(getfoo, 'bar', doc='My bardoc.')"

        # name
        inst = self.klass(name='bar')
        assert repr(inst) == "PropBase(None, 'bar')"

        # name + default + doc
        inst = self.klass(name='bar', default='Baz', doc='My bardoc.')
        assert repr(inst) == \
            "PropBase(None, 'bar', default='Baz', doc='My bardoc.')"

    def test_get(self):
        """
        Test the `state.PropBase.__get__` method.
        """
        class Example(Dummy):
            @self.klass
            def row(self):
                self.i += 1
                return 'row %d' % self.i

            tag = self.klass(name='tag', default='Tag')

        eg = Example()
        assert eg.row == 'row 1'
        assert eg.row == 'row 2'

        assert eg.tag == 'Tag'
        eg._tag = "you're it"
        assert eg.tag == "you're it"

    def test_set(self):
        """
        Test the `state.PropBase.__set__` method.
        """
        class Foo(Dummy):
            @self.klass
            def foo(self):
                return 'FOO'

        f = Foo()
        e = raises(AttributeError, setattr, f, 'foo', 'nope')
        assert str(e) == "read-only PropBase: cannot set Foo.foo to 'nope'"

    def test_delete(self):
        """
        Test the `state.PropBase.__delete__` method.
        """
        class Bar(Dummy):
            @self.klass
            def bar(self):
                return 'BAR'

        b = Bar()
        e = raises(AttributeError, delattr, b, 'bar')
        assert str(e) == 'read-only PropBase: cannot del Bar.bar'


class test_StaticProp(object):
    """
    Test the `state.StaticProp` class.
    """

    klass = state.StaticProp

    def test_set(self):
        """
        Test the `state.StaticProp.__set__` method.
        """
        class Foo(Dummy):
            @self.klass
            def foo(self):
                return 'FOO'

        f = Foo()
        e = raises(AttributeError, setattr, f, 'foo', 'nope')
        assert str(e) == "read-only StaticProp: cannot set Foo.foo to 'nope'"

    def test_delete(self):
        """
        Test the `state.StaticProp.__delete__` method.
        """
        class Bar(Dummy):
            @self.klass
            def bar(self):
                return 'BAR'

        b = Bar()
        e = raises(AttributeError, delattr, b, 'bar')
        assert str(e) == 'read-only StaticProp: cannot del Bar.bar'


class test_DynamicProp(object):
    """
    Test the `state.DynamicProp` class.
    """

    klass = state.DynamicProp

    def test_set(self):
        """
        Test the `state.DynamicProp.__set__` method.
        """
        class Foo(Dummy):
            @self.klass
            def foo(self):
                return 'FOO'

        f = Foo()
        e = raises(AttributeError, setattr, f, 'foo', 'nope')
        assert str(e) == "read-only DynamicProp: cannot set Foo.foo to 'nope'"

    def test_delete(self):
        """
        Test the `state.DynamicProp.__delete__` method.
        """
        class Bar(Dummy):
            @self.klass
            def bar(self):
                return 'BAR'

        b = Bar()
        e = raises(AttributeError, delattr, b, 'bar')
        assert str(e) == 'read-only DynamicProp: cannot del Bar.bar'


class test_Statefull(object):
    """
    Test the `state.Statefull` class.
    """

    klass = state.Statefull

    def example_cls(self):
        class Example(self.klass):
            name = state.Static('name', default='foo')
            id = state.Static('id')

            title = state.Dynamic('title', default='Gone with the wind')
            label = state.Dynamic('label')

            @state.StaticProp
            def tag(self):
                return 'Tag'
            @state.StaticProp
            def children(self):
                return ('child1', 'child2')
            @state.StaticProp
            def klass(self):
                return self.__class__.__name__

            @state.DynamicProp
            def css_classes(self):
                return (self.name, self.title)

        names = (
            'Static',
            'Dynamic',
            'StaticProp',
            'DynamicProp',
        )
        for name in names:
            t = getattr(Example, name)
            assert isinstance(t, tuple)

        return Example

    def example_inst(self):
        Example = self.example_cls()
        inst = Example()
        return inst

    def test_state(self):
        """
        Test the `state.Statefull.state` method.
        """
        inst = self.example_inst()
        state = inst.state()
        assert state == dict(
            name='foo',
            id=None,
            title='Gone with the wind',
            label=None,
            tag='Tag',
            children=('child1', 'child2'),
            klass='Example',
            css_classes=('foo', 'Gone with the wind'),
        )
        assert inst.state() == state  # Call it again
        assert inst.__state__ is None
        assert inst.__request__ == dict()

        # freeze it
        freeze(inst)
        assert inst.state() == state
        no_DynamicProp = dict(state)
        del no_DynamicProp['css_classes']
        assert type(inst.__state__) is tuple
        assert dict(inst.__state__) == no_DynamicProp
        assert inst.__request__ == dict()

        # set a request
        request = dict(
            title='Dirty Harry',
            label='wins another Oscar',
        )
        inst.__request__.update(request)
        state2 = dict(state)
        state2.update(request)
        state2['css_classes'] = ('foo', 'Dirty Harry')
        assert inst.state() == state2
        assert inst.state() != state
        assert inst.state() == state2  # Check again
        assert dict(inst.__state__) == no_DynamicProp
        assert inst.__request__ == request

        # clear the request and check again
        inst.clear()
        assert inst.__request__ == dict()
        assert inst.state() == state
        assert dict(inst.__state__) == no_DynamicProp

    def test_request(self):
        """
        Test the `state.Statefull.request` property.
        """
        inst = self.example_inst()
        orig = inst.state()
        dynamic = dict(label=None, title='Gone with the wind')
        assert inst.request() == dynamic
        assert inst.__request__ == dict()

        # We'll try to override everything
        request = dict(
            name='The name',
            id='The id',
            title='The title',
            label='The label',
            tag='The tag',
            children='The children',
            klass='The klass',
            css_classes='The css_classes',
        )
        expected = dict(
            title='The title',
            label='The label',
        )

        # But nothing should happen if not frozen:
        inst.update_request(request)
        assert inst.request() == dynamic
        assert inst.__request__ == dict()
        assert inst.state() == orig

        # And even after freeze(), we should only be able to set `Dynamic`:
        freeze(inst)
        inst.update_request(request)
        assert inst.request() == expected
        assert inst.__request__ == expected

        # So the new state should only have `Dynamic` changed and possibly
        # `DynamicProp` that are driven by `Dynamic`:
        new = dict(orig)
        new.update(expected)
        new['css_classes'] = ('foo', 'The title')
        assert inst.state() == new

        # Clear the request and make sure the state goes back to orig:
        inst.clear()
        assert inst.request() == dynamic
        assert inst.__request__ == dict()
        assert inst.state() == orig

    def test_clear(self):
        """
        Test the `state.Statefull.clear` method.
        """
        inst = self.example_inst()
        assert inst.__request__ == dict()
        inst.__request__.update(
            dict(
                aye='a',
                bee='b',
            )
        )
        assert inst.__request__ == dict(aye='a', bee='b')
        inst.clear()
        assert inst.__request__ == dict()
