# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the `wehjit.styles` module.
"""

from helpers import raises
from wehjit import styles, errors


def test_dict_diff():
    """
    Test the `styles.dict_diff` function.
    """
    f = styles.dict_diff

    # Test at depth=0:
    a0 = dict(
        one='One',
        two=2,
        four=4,
    )
    b0 = dict(
        one=1,
        two=2,
        three=3,
    )
    assert dict(f(None, None)) == dict()
    assert dict(f(None, b0)) == dict()
    assert dict(f(a0, None)) == a0
    assert dict(f(a0, b0)) == dict(
        one='One',
        four=4,
    )

    # Test at depth=1:
    a1 = dict(
        aye=a0,
        bee=a0,
    )
    b1 = dict(
        aye=a0,
        bee=b0,
    )
    assert dict(f(None, None, depth=1)) == dict()
    assert dict(f(None, b1, depth=1)) == dict()
    assert dict(f(a1, None, depth=1)) == a1
    assert dict(f(a1, b1, depth=1)) == dict(
        bee=dict(
            one='One',
            four=4,
        ),
    )


def test_filter_by_value():
    """
    Test the `styles.filter_by_value` function.
    """
    f = styles.filter_by_value
    assert tuple(f(None)) == tuple()
    assert tuple(f({})) == tuple()

    # Test at depth=0:
    d0 = {
        'text-align': 'center',
        'color': None,
        'background': '#f00',
        'border': '',
        'font-size': '11pt',
    }
    assert tuple(f(d0)) == (
        ('background', '#f00'),
        ('font-size', '11pt'),
        ('text-align', 'center'),
    )

    # Test at depth=1:
    d1 = {
        '.current': {
            'background': '%(current)s',
            'color': '',
        },
        '': d0,
        '> a': {
            'text-decoration': '',
            'border': None,
        },
    }
    assert tuple(f(d1, depth=1)) == (
        (
            '',
            (
                ('background', '#f00'),
                ('font-size', '11pt'),
                ('text-align', 'center'),
            ),
        ),
        (
            '.current',
            (
                ('background', '%(current)s'),
            ),
        ),
    )


class test_Styles(object):
    """
    Test the `styles.Styles` class.
    """
    klass = styles.Styles

    def test_loads(self):
        """
        Test the `styles.Styles.loads` method.
        """
        s = self.klass()
        e = raises(ValueError, s.loads, 'foo bar')
        assert str(e) == 'No JSON object could be decoded'
        e = raises(TypeError, s.loads, '[1, 2, 3]')
        assert str(e) == 'JSON data must define %r; got a %r: %r' % (
            dict, list, [1, 2, 3]
        )

    def test_substitute(self):
        """
        Test the `styles.Styles._substitute` method.
        """
        inst = self.klass()
        ns = 'Menu'
        s = '> ul'
        d = {
            'background': '%(menu_background)s',
            'border': '%(b_width)s solid %(b_color)s',
            'color': '#000',
        }
        inst.update(ns, s, d)
        inst.update_vars(
            menu_background='#ddd',
            b_width='1px',
        )

        # Test with undefined style var
        gen = inst._substitute(ns, s)
        e = raises(errors.StyleVarError, dict, gen)
        assert e.var == 'b_color'
        assert e.key == 'border'
        assert e.value == '%(b_width)s solid %(b_color)s'
        assert e.namespace == ns
        assert e.selector == s

        # Test with all vars defined:
        inst.update_vars(
            b_color='#666',
        )
        gen = inst._substitute(ns, s)
        assert dict(gen) == {
            'background': '#ddd',
            'border': '1px solid #666',
            'color': '#000',
        }

        # Test with non-existent namespace:
        assert dict(inst._substitute('Foo', s)) == dict()

        # Test with non-existent selector:
        assert dict(inst._substitute(ns, '> a.bar')) == dict()
